// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-core/agent.h>
#include "channel_group.h"

using namespace yami;
using namespace core;

agent::agent()
    : initialized_(false), alloc_(&default_alloc_),
      ch_group_(NULL),
      max_channels_allowed_(0), channels_used_(0) // dummy initialization
{
}

result agent::init(
    incoming_message_dispatch_function dispatch_callback,
    void * dispatch_hint,
    closed_connection_function disconnection_hook,
    void * disconnection_hook_hint,
    allocator * alloc)
{
    result res;
    if (initialized_)
    {
        res = bad_state;
    }
    else
    {
        res = do_init(NULL,
            dispatch_callback, dispatch_hint,
            disconnection_hook, disconnection_hook_hint,
            alloc);
    }

    return res;
}

result agent::init(const parameters & configuration_options,
    incoming_message_dispatch_function dispatch_callback,
    void * dispatch_hint,
    closed_connection_function disconnection_hook,
    void * disconnection_hook_hint,
    allocator * alloc)
{
    result res;
    if (initialized_)
    {
        res = bad_state;
    }
    else
    {
        res = do_init(&configuration_options,
            dispatch_callback, dispatch_hint,
            disconnection_hook, disconnection_hook_hint,
            alloc);
    }

    return res;
}

result agent::do_init(const parameters * configuration_options,
    incoming_message_dispatch_function dispatch_callback,
    void * dispatch_hint,
    closed_connection_function disconnection_hook,
    void * disconnection_hook_hint,
    allocator * alloc)
{
    if (alloc != NULL)
    {
        alloc_ = alloc;
    }

    // allocate the implementation object
    ch_group_ = static_cast<details::channel_group *>(
        alloc_->allocate(sizeof(details::channel_group)));

    core::result res;
    if (ch_group_ != NULL)
    {
        res = ch_group_->init(*alloc_, configuration_options,
            dispatch_callback, dispatch_hint,
            disconnection_hook, disconnection_hook_hint);

        if (res == core::ok)
        {
            initialized_ = true;
        }
        else
        {
            alloc_->deallocate(ch_group_);
        }
    }
    else
    {
        res = no_memory;
    }

    return res;
}

void agent::install_event_notifications(
    event_notification_function event_notification_callback,
    void * event_notification_hint)
{
    ch_group_->install_event_notifications(
        event_notification_callback, event_notification_hint);
}

void agent::install_frame_acceptor(
    frame_acceptor_function frame_acceptor,
    void * frame_acceptor_hint)
{
    ch_group_->install_frame_acceptor(
        frame_acceptor, frame_acceptor_hint);
}

void agent::install_io_error_logger(
    io_error_function io_error_callback,
    void * io_error_callback_hint)
{
    ch_group_->install_io_error_logger(
        io_error_callback, io_error_callback_hint);
}

void agent::clean()
{
    if (initialized_)
    {
        ch_group_->clean();

        alloc_->deallocate(ch_group_);

        initialized_ = false;
    }
}

agent::~agent()
{
    clean();
}

result agent::open(const char * target)
{
    channel_descriptor dummy_descriptor;
    bool dummy_created_new_channel;
    return open(target, dummy_descriptor, dummy_created_new_channel);
}

result agent::open(const char * target, channel_descriptor & new_channel,
    bool & created_new_channel)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->open(target, new_channel, created_new_channel);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::open(const char * target, channel_descriptor & new_channel,
    bool & created_new_channel, const parameters * overriding_options)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->open(target, new_channel, created_new_channel,
            true, overriding_options);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::is_open(const char * target,
    channel_descriptor & existing_channel) const
{
    result res;
    if (initialized_)
    {
        res = ch_group_->is_open(target, existing_channel);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::close(channel_descriptor cd, std::size_t priority)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->close(cd, false, priority);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::close(const char * target, std::size_t priority)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->close(target, false, priority);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::hard_close(channel_descriptor cd)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->close(cd, true, 0);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::hard_close(const char * target)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->close(target, true, 0);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::post(channel_descriptor cd,
    const serializable & message_header,
    const serializable & message_body,
    std::size_t priority,
    message_progress_function progress_callback,
    void * progress_hint)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->post(cd, message_header, message_body, priority,
            progress_callback, progress_hint);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::post(const char * target,
    const serializable & message_header,
    const serializable & message_body,
    std::size_t priority,
    message_progress_function progress_callback,
    void * progress_hint)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->post(target, message_header, message_body, priority,
            progress_callback, progress_hint);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::add_listener(const char * target,
    new_incoming_connection_function connection_hook,
    void * connection_hook_hint,
    const char * * resolved_target)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->add_listener(target,
            connection_hook, connection_hook_hint,
            resolved_target);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::remove_listener(const char * target)
{
    result res;
    if (initialized_)
    {
        ch_group_->remove_listener(target);
        res = ok;
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::do_some_work(std::size_t timeout,
    bool allow_outgoing_traffic, bool allow_incoming_traffic)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->do_some_work(timeout,
            allow_outgoing_traffic, allow_incoming_traffic);

        if (res == ok)
        {
            // only now the selector has up-to-date usage values
            // -> store them aside for future requests
            
            ch_group_->get_channel_usage(
                max_channels_allowed_, channels_used_);
        }
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::interrupt_work_waiter()
{
    result res;
    if (initialized_)
    {
        res = ch_group_->interrupt_work_waiter();
    }
    else
    {
        res = bad_state;
    }

    return res;
}

void agent::get_channel_usage(std::size_t & max_allowed, std::size_t & used)
{
    max_allowed = max_channels_allowed_;
    used = channels_used_;
}

result agent::get_pending_outgoing_bytes(
    channel_descriptor cd, std::size_t & bytes)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->get_pending_outgoing_bytes(cd, bytes);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

result agent::get_pending_outgoing_bytes(
    const char * target, std::size_t & bytes)
{
    result res;
    if (initialized_)
    {
        res = ch_group_->get_pending_outgoing_bytes(target, bytes);
    }
    else
    {
        res = bad_state;
    }

    return res;
}

