// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#ifndef YAMICORE_CHANNEL_GROUP_H_INCLUDED
#define YAMICORE_CHANNEL_GROUP_H_INCLUDED

#include "channel.h"
#include <yami4-core/channel_descriptor.h>
#include <yami4-core/core.h>
#include "details-fwd.h"
#include "options.h"

#ifdef YAMI4_WITH_OPEN_SSL
#ifdef _WIN32
// this is necessary to avoid conflicts between
// WinSock2.h (used by YAMI4) and winsock.h (used by OpenSSL):
#include <WinSock2.h>
#endif // _WIN32
#include <openssl/ssl.h>
#endif // YAMI4_WITH_OPEN_SSL

// selected per platform
#include <yami4-sys/details-types.h>
#include <yami4-sys/mutex.h>
#include <yami4-sys/selector.h>

namespace yami
{

namespace details
{

class channel_group
{
public:

    core::result init(core::allocator & alloc,
        const core::parameters * configuration_options,
        core::incoming_message_dispatch_function dispatch_callback,
        void * dispatch_hint,
        core::closed_connection_function disconnection_hook,
        void * disconnection_hook_hint);

    void install_event_notifications(
        core::event_notification_function event_notification_callback,
        void * event_notification_hint);

    void install_frame_acceptor(
        core::frame_acceptor_function frame_acceptor,
        void * frame_acceptor_hint);

    void install_io_error_logger(
        core::io_error_function io_error_callback,
        void * io_error_callback_hint);

    void clean();

    // opens new channel for the given target
    // or returns the descriptor of the existing channel
    core::result open(const char * target,
        core::channel_descriptor & new_channel,
        bool & created_new_channel, bool do_lock = true,
        const core::parameters * overriding_options = NULL);

    // opens new channel for the given target,
    // and reports it as incoming connection (used for datagram protocols)
    core::result open_as_incoming(const char * target,
        core::channel_descriptor & new_channel,
        bool & created_new_channel, bool do_lock,
        const core::parameters * overriding_options);

    core::result is_open(const char * target,
        core::channel_descriptor & existing_channel) const;

    // adds an existing channel to the list, used to add new channels
    // created by listeners (fd is already valid);
    // target is copied
    core::result add_existing(
        const char * target, io_descriptor_type & fd, protocol prot,
        std::size_t preferred_frame_size,
        core::channel_descriptor & new_descriptor,
        channel * & new_channel);
#ifdef YAMI4_WITH_OPEN_SSL
    core::result add_existing_ssl(
        const char * target, const io_descriptor_type & fd, protocol prot,
        std::size_t preferred_frame_size,
        core::channel_descriptor & new_descriptor);
#endif // YAMI4_WITH_OPEN_SSL

    // insert a poison pill to the given channel
    // do nothing if no such channel exists
    // close immediately and ignore priority if hard_close == true
    core::result close(core::channel_descriptor cd,
        bool hard_close, std::size_t priority);
    core::result close(const char * target,
        bool hard_close, std::size_t priority);

    // serializes the new message into frames and injects them
    // into the output queue of the given channel
    core::result post(core::channel_descriptor cd,
        const core::serializable & message_header,
        const core::serializable & message_body,
        std::size_t priority,
        core::message_progress_function progress_callback,
        void * progress_hint);
    core::result post(const char * target,
        const core::serializable & message_header,
        const core::serializable & message_body,
        std::size_t priority,
        core::message_progress_function progress_callback,
        void * progress_hint);

    core::result add_listener(const char * target,
        core::new_incoming_connection_function connection_hook,
        void * connection_hook_hint,
        const char * * resolved_target);
    void remove_listener(const char * target);

    // waits for the readiness of any dependent channel
    // to perform I/O work and executing a unit of work
    // (the unit of work can involve many channels, if more than one
    // is ready for transmission; it can also return without
    // physically doing any I/O-related work)
    // the timeout is in milliseconds,
    // 0 means immediate check and action (non-blocking)
    core::result do_some_work(std::size_t timeout,
        bool allow_outgoing_traffic,
        bool allow_incoming_traffic);

    // used to inject already received frames (UDP),
    // should be called from synchronized context
    core::result process_complete_incoming_frame(core::channel_descriptor cd,
        const char * buffer, const std::size_t buffer_size);

#ifdef YAMI4_WITH_QNX
    // used to establish the channel id for receiving responses
    // from remote peers, this id is sent together with each outgoing message;
    // only the first created QNX listener is remembered
    void set_default_qnx_listening_channel_id(int chid);
#endif // YAMI4_WITH_QNX

    core::result interrupt_work_waiter();

    const options & get_options() const { return configuration_options_; }

    void get_channel_usage(std::size_t & max_allowed, std::size_t & used);
    
    core::result get_pending_outgoing_bytes(core::channel_descriptor cd, std::size_t & bytes);
    core::result get_pending_outgoing_bytes(const char * target, std::size_t & bytes);
    
    // for unit tests
    channel_group() {}
    
private:
    
    channel_group(const channel_group &);
    void operator=(const channel_group &);

    core::result do_open(const char * target,
        core::channel_descriptor & new_channel,
        bool & created_new_channel, bool do_lock,
        const core::parameters * overriding_options,
        core::event_notification event);
    core::result do_close(channel * ch, std::size_t priority,
        std::size_t index);
    core::result do_hard_close(channel * ch, std::size_t index);

    // finds existing channel with the given target name
    // returns ok or no_such_name
    core::result find_existing_channel(const char * target,
        std::size_t & index, std::size_t & sequence_number,
        channel * & ch) const;

    // finds an empty slot or makes room for new empty slot
    // reservation means that the empty slot will not be found
    // the second time until it is filled and cleared for reuse
    // returns ok or no_memory
    core::result find_unused_channel(std::size_t & index, bool reserve);

    bool channel_dec_ref(std::size_t index, channel * ch);

    void prune_listeners();

    std::size_t generate_message_id();

    mutable mutex mtx_;

    options configuration_options_;

    channel_holder * channel_holders_;
    std::size_t channels_num_;
    channel * * shadow_channels_;
    std::size_t shadow_channels_num_;

    listener * first_listener_;

#ifdef YAMI4_WITH_QNX
    int qnx_listener_channel_id_;
#endif // YAMI4_WITH_QNX

    selector selector_;

    core::allocator * alloc_;

    bool closing_;

    std::size_t last_message_id_;

    core::incoming_message_dispatch_function incoming_message_callback_;
    void * incoming_message_hint_;
    core::closed_connection_function disconnection_hook_;
    void * disconnection_hook_hint_;

    core::event_notification_function event_notification_callback_;
    void * event_notification_hint_;

    core::frame_acceptor_function frame_acceptor_callback_;
    void * frame_acceptor_hint_;

    core::io_error_function io_error_callback_;
    void * io_error_callback_hint_;

#ifdef YAMI4_WITH_OPEN_SSL
    SSL_CTX * ssl_ctx_;
#endif // YAMI4_WITH_OPEN_SSL
};

} // namespace details

} // namespace yami

#endif // YAMICORE_CHANNEL_GROUP_H_INCLUDED
