// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-core/agent.h>
#include <yami4-core/locked_allocator.h>
#include <yami4-core/parameters.h>

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>

#include <atomic>
#include <thread>

using namespace yami::core;

std::atomic_bool worker_can_stop(false);

// worker thread function that is used to process I/O events
// for the client agent
// this thread should stop working before the agent is closed
extern "C" void worker_thread_function(agent * client_agent)
{
    const std::size_t time_out = 1000; // 1 second

    while (worker_can_stop.load() == false)
    {
        result res = client_agent->do_some_work(time_out);
        if (res == no_memory)
        {
            std::cout << "*** not enough memory for doing the work\n";
        }
        
        assert(res == ok || res == timed_out);
    }
}

void run_client(const char * server_address, void * memory, std::size_t memory_size)
{
    locked_allocator alloc;
    alloc.set_working_area(memory, memory_size);

    agent client_agent;

    // create the client agent
    // and open a channel to the given server

    result res = client_agent.init(NULL, NULL, NULL, NULL, &alloc);
    assert(res == ok);

    res = client_agent.open(server_address);
    if (res == no_memory)
    {
        std::cout << "*** not enough memory to open connection\n";
    }
    
    assert(res == ok);

    // start the worker thread that will process I/O events

    std::thread worker_thread(worker_thread_function, &client_agent);

    // read lines of text from standard input
    // and post each one for transmission

    std::string input_line;
    
    while (std::getline(std::cin, input_line))
    {
        parameters empty_header;
        parameters body;

        // the "content" field name is arbitrary,
        // but needs to be recognized at the server side

        res = body.set_string(
            "content", input_line.c_str());
        assert(res == ok);

        res = client_agent.post(
            server_address, empty_header, body);
        if (res == no_memory)
        {
            std::cout << "*** no memory, skipping the last line\n";
        }
        else
        {
            assert(res == ok);
        }
    }

    // stop the worker thread
    // and clean up the client resources

    worker_can_stop.store(true);
    worker_thread.join();

    // the cleanup is automatic in the agent's destructor
}

int main(int argc, char * argv[])
{
    if (argc != 3)
    {
        std::cout
            << "expecting two parameters: server destination, memory size\n";
        return EXIT_FAILURE;
    }

    const char * server_address = argv[1];
    std::size_t memory_size = std::stoi(argv[2]);

    std::vector<char> memory(memory_size);

    run_client(server_address, &memory[0], memory_size);
}
