// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include "listener.h"
#include <yami4-core/allocator.h>
#include <cstring>

using namespace yami;
using namespace details;

namespace // unnamed
{
const char tcp_prefix[] = "tcp://";
const char udp_prefix[] = "udp://";
#if !defined(_WIN32)
const char unix_prefix[] = "unix://";
#endif // !_WIN32

#ifdef YAMI4_WITH_QNX
const char qnx_prefix[] = "qnx://";
#endif // YAMI4_WITH_QNX

#ifdef YAMI4_WITH_OPEN_SSL
const char tcps_prefix[] = "tcps://";
#endif // YAMI4_WITH_OPEN_SSL
} // unnamed namespace

void listener::init(core::allocator & alloc,
    channel_group & group,
    core::new_incoming_connection_function connection_hook,
    void * connection_hook_hint,
    core::io_error_function io_error_callback,
    void * io_error_callback_hint)
{
    next = NULL;
    alloc_ = &alloc;
    target_ = NULL;
    ref_count_ = 1;
    tcp_port_ = -1;
    fd_ = empty_io_descriptor;
    selector_index_ = 0;
    group_ = &group;
    frame_buffer_ = NULL;
    connection_hook_ = connection_hook;
    connection_hook_hint_ = connection_hook_hint;
    io_error_callback_ = io_error_callback;
    io_error_callback_hint_ = io_error_callback_hint;
}

core::result listener::prepare(const char * target)
{
    core::result res;

    // recognize the protocol

    const std::size_t tcp_prefix_size = sizeof(tcp_prefix) - 1;
    const std::size_t udp_prefix_size = sizeof(udp_prefix) - 1;
#if !defined(_WIN32)
    const std::size_t unix_prefix_size = sizeof(unix_prefix) - 1;
#endif // !_WIN32
#ifdef YAMI4_WITH_QNX
    const std::size_t qnx_prefix_size = sizeof(qnx_prefix) - 1;
#endif // YAMI4_WITH_QNX
#ifdef YAMI4_WITH_OPEN_SSL
    const std::size_t tcps_prefix_size = sizeof(tcps_prefix) - 1;
#endif // YAMI4_WITH_OPEN_SSL

    if (std::strncmp(target, tcp_prefix, tcp_prefix_size) == 0)
    {
        res = prepare_tcp(target + tcp_prefix_size);
    }
    else if (std::strncmp(target, udp_prefix, udp_prefix_size) == 0)
    {
        res = prepare_udp(target + udp_prefix_size);
    }
#if !defined(_WIN32)
    else if (std::strncmp(target, unix_prefix, unix_prefix_size) == 0)
    {
        res = prepare_unix(target + unix_prefix_size);
    }
#endif // !_WIN32
#ifdef YAMI4_WITH_QNX
    else if (std::strncmp(target, qnx_prefix, qnx_prefix_size) == 0)
    {
        res = prepare_qnx();
    }
#endif // YAMI4_WITH_QNX
#ifdef YAMI4_WITH_OPEN_SSL
    else if (std::strncmp(target, tcps_prefix, tcps_prefix_size) == 0)
    {
        res = prepare_tcps(target + tcps_prefix_size);
    }
#endif // YAMI4_WITH_OPEN_SSL
    else
    {
        // unknown protocol

        res = core::bad_protocol;
    }

    return res;
}

// synchronized by caller
void listener::clean()
{
    close_resource();
    alloc_->deallocate(target_);

    if (frame_buffer_ != NULL)
    {
        alloc_->deallocate(frame_buffer_);
    }
}
