// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#ifndef YAMICORE_LISTENER_H_INCLUDED
#define YAMICORE_LISTENER_H_INCLUDED

#include <yami4-core/core.h>
#include "details-fwd.h"
#include <cstddef>

#ifdef YAMI4_WITH_OPEN_SSL
#ifdef _WIN32
// this is necessary to avoid conflicts between
// WinSock2.h (used by YAMI4) and winsock.h (used by OpenSSL):
#include <WinSock2.h>
#endif // _WIN32
#include <openssl/ssl.h>
#endif // YAMI4_WITH_OPEN_SSL

// selected per platform
#include <yami4-sys/details-types.h>

namespace yami
{

namespace details
{

class listener
{
public:

    void init(core::allocator & alloc,
        channel_group & group,
        core::new_incoming_connection_function connection_hook,
        void * connection_hook_hint,
        core::io_error_function io_error_callback,
        void * io_error_callback_hint);

    core::result prepare(const char * target);

    core::result do_some_work();

    void close_resource();
    void clean();

    const char * get_target() const { return target_; }
    protocol get_protocol() const { return protocol_; }

    const io_descriptor_type & get_io_descriptor() const { return fd_; }

    void set_selector_index(std::size_t index) { selector_index_ = index; }
    std::size_t get_selector_index() const { return selector_index_; }
    
#ifdef YAMI4_WITH_QNX
    bool is_qnx_receiver_ready() const;
#endif // YAMI4_WITH_QNX

    void inc_ref() { ++ref_count_; }
    void dec_ref() { --ref_count_; }
    bool can_be_removed() const { return ref_count_ == 0; }

    listener * next;

private:
    
    listener(const listener &);
    void operator=(const listener &);

    protocol protocol_;

    core::result prepare_tcp(const char * address);
    core::result prepare_udp(const char * address);
#if !defined(_WIN32)
    core::result prepare_unix(const char * path);
#endif // !_WIN32
    
#ifdef YAMI4_WITH_QNX
    core::result prepare_qnx();
#endif // YAMI4_WITH_QNX

    core::result accept_tcp();
    core::result accept_udp();
#if !defined(_WIN32)
    core::result accept_unix();
#endif // !_WIN32

#ifdef YAMI4_WITH_QNX
    core::result accept_qnx();
#endif // YAMI4_WITH_QNX
    
#ifdef YAMI4_WITH_OPEN_SSL
    core::result prepare_tcps(const char * address);
    core::result accept_tcps();
#endif // YAMI4_WITH_OPEN_SSL

    core::allocator * alloc_;
    
    std::size_t ref_count_;

    const char * target_;

    int tcp_port_; // network port (in local order)

    io_descriptor_type fd_;
    
    std::size_t selector_index_;

    channel_group * group_;

    // used only for UDP
    char * frame_buffer_;
    std::size_t frame_buffer_size_;
    
#ifdef YAMI4_WITH_QNX
    int channel_id_;
    msg_receiver msg_receiver_;
#endif // YAMI4_WITH_QNX

    core::new_incoming_connection_function connection_hook_;
    void * connection_hook_hint_;

    core::io_error_function io_error_callback_;
    void * io_error_callback_hint_;
};

} // namespace details

} // namespace yami

#endif // YAMICORE_LISTENER_H_INCLUDED
