// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#ifndef YAMICORE_PARAMETERS_DETAILS_H_INCLUDED
#define YAMICORE_PARAMETERS_DETAILS_H_INCLUDED

#include <yami4-core/core.h>
#include "details-fwd.h"
#include <yami4-core/parameter_type.h>
#include <cstddef>

namespace yami
{

namespace details
{

const std::size_t short_name_optimization_threshold = 16u;
const std::size_t initial_number_of_entries = 4u;
const std::size_t max_nesting_level = 5u;

struct entry_name
{
    std::size_t name_length;
    union packed
    {
        const char * long_value;
        char short_value[short_name_optimization_threshold];
    } buffer;

    core::result set(const char * value, std::size_t length,
        core::allocator & alloc);

    const char * value() const;

    void clear(core::allocator & alloc);

    bool equals(const char * value, std::size_t length) const;

    // for unit tests
    void dump(details::dump_sink & sink) const;
};

struct string_array_element
{
    const char * value;
    std::size_t length;

    void clear(core::allocator & alloc);
};

struct binary_array_element
{
    const void * value;
    std::size_t length;

    void clear(core::allocator & alloc);
};

struct entry
{
    entry_name name;

    core::parameter_type type;

    union packed
    {
        bool b;
        int i;
        long long L;
        unsigned int ui;
        unsigned long long uL;
        double d;
        struct string_data
        {
            const char * value;
            std::size_t length;
            bool own;
        } str;
        struct binary_data
        {
            const void * value;
            std::size_t length;
            bool own;
        } bin;
        struct boolean_array_data
        {
            bool * values;
            std::size_t length;
            bool own;
        } ba;
        struct integer_array_data
        {
            int * values;
            std::size_t length;
            bool own;
        } ia;
        struct long_long_array_data
        {
            long long * values;
            std::size_t length;
            bool own;
        } La;
        struct unsigned_integer_array_data
        {
            unsigned int * values;
            std::size_t length;
            bool own;
        } uia;
        struct unsigned_long_long_array_data
        {
            unsigned long long * values;
            std::size_t length;
            bool own;
        } uLa;
        struct double_float_array_data
        {
            double * values;
            std::size_t length;
            bool own;
        } da;
        struct string_array_data
        {
            string_array_element * values;
            std::size_t length;
            bool own;
        } sa;
        struct binary_array_data
        {
            binary_array_element * values;
            std::size_t length;
            bool own;
        } bina;
        core::parameters * nested;
        struct nested_array_data
        {
            core::parameters * values;
            std::size_t length;
            bool own;
        } nesteda;
    } item;

    core::result set_name(const char * value, std::size_t length,
        core::allocator & alloc)
    {
        return name.set(value, length, alloc);
    }

    void clear_name(core::allocator & alloc)
    {
        name.clear(alloc);
    }

    bool name_equals(const char * value, std::size_t length) const
    {
        return name.equals(value, length);
    }

    void clear_item(core::allocator & alloc);
};

void get_serialize_buffer_size(
    const core::parameters & params,
    std::size_t & size); // starts with 0 (for root) and gets updated

core::result serialize_entry(
    const details::entry & e,
	char * * buffers,
    const std::size_t * buffer_sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position);

core::result serialize(
    const core::parameters & params,
	char * * buffers,
    const std::size_t * buffer_sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position);

core::result deserialize_entry(
    core::parameters & params,
	const char * * buffers,
    const std::size_t * buffer_sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position);

core::result deserialize(
    core::parameters & params,
	const char * * buffers,
    const std::size_t * buffer_sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position);

int type_code(core::parameter_type t);

core::result get_type_from_code(int code, core::parameter_type & type);

// find the entry with given name
// returns the index or num_of_entries if not found
std::size_t find_entry(const entry * data, std::size_t num_of_entries,
    const char * name, std::size_t name_length);

// find an appropriate place for a new entry
// by either locating an unused entry
// or by extending the set
core::result find_empty_entry(entry * & data, std::size_t & num_of_entries,
    std::size_t & index, core::allocator & alloc);

// finds an existing entry with the given name and clears its item
// or creates a new entry with this name already set
core::result prepare_for_set(entry * & data, std::size_t & num_of_entries,
    const char * name, std::size_t name_length,
    std::size_t & index, core::allocator & alloc);

// finds the next used entry or returns num_of_entries if nothing was found
std::size_t find_next_used(const entry * data, std::size_t num_of_entries,
    std::size_t current_index);

core::result do_set_string(const char * name, std::size_t name_length,
    const char * value, std::size_t value_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_binary(const char * name, std::size_t name_length,
    const void * value, std::size_t value_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_boolean_array(const char * name, std::size_t name_length,
    const bool * values, std::size_t array_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_integer_array(const char * name, std::size_t name_length,
    const int * values, std::size_t array_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_long_long_array(
    const char * name, std::size_t name_length,
    const long long * values, std::size_t array_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_word_array(const char * name, std::size_t name_length,
    const unsigned int * values, std::size_t array_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_long_long_word_array(
    const char * name, std::size_t name_length,
    const unsigned long long * values, std::size_t array_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_set_double_float_array(
    const char * name, std::size_t name_length,
    const double * values, std::size_t array_length,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc, bool own);

core::result do_create_string_array(
    const char * name, std::size_t name_length,
    std::size_t array_length, std::size_t & index,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc);

core::result do_set_string_in_array(
    std::size_t item_index, std::size_t array_index,
    const char * value, std::size_t value_length,
    entry * data, core::allocator & alloc);

core::result do_place_string_in_array(
    std::size_t item_index, std::size_t array_index,
    const char * value, std::size_t value_length,
    entry * data, core::allocator & alloc);

core::result do_create_binary_array(
    const char * name, std::size_t name_length,
    std::size_t array_length, std::size_t & index,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc);

core::result do_set_binary_in_array(
    std::size_t item_index, std::size_t array_index,
    const void * value, std::size_t value_length,
    entry * data, core::allocator & alloc);

core::result do_place_binary_in_array(
    std::size_t item_index, std::size_t array_index,
    const void * value, std::size_t value_length,
    entry * data, core::allocator & alloc);

core::result do_create_nested_array(
    const char * name, std::size_t name_length,
    std::size_t array_length, std::size_t & index,
    entry * & data, std::size_t & num_of_entries,
    core::allocator & alloc);

core::result do_access_nested_in_array(
    std::size_t item_index, std::size_t array_index,
    core::parameters * & nested,
    entry * data);

} // namespace details

} // namespace yami

#endif // YAMICORE_PARAMETERS_DETAILS_H_INCLUDED
