// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include "selector.h"
#include <yami4-core/allocator.h>
#include "../../channel.h"
#include "../../io_error_handler.h"
#include "../../listener.h"
#include <time.h>

#include <cstring>
#include <errno.h>

using namespace yami;
using namespace details;

core::result selector::init(core::allocator & alloc)
{
    io_error_callback_ = NULL;

    alloc_ = &alloc;
    polls_ = NULL;

    core::result res;

    polls_size_ = 10; // initial size of poll array
    polls_ = static_cast<pollfd *>(
        alloc_->allocate(sizeof(struct pollfd) * polls_size_));
    if (polls_ != NULL)
    {
        polls_used_ = 0;

        io_descriptor_type pipe_descriptors[2];

        int cc = ::pipe(pipe_descriptors);
        if (cc == 0)
        {
            interrupt_pipe_write_ = pipe_descriptors[1];
            interrupt_pipe_read_ = pipe_descriptors[0];

            res = core::ok;
        }
        else
        {
            alloc_->deallocate(polls_);
            polls_ = NULL;

            res = core::io_error;
        }
    }
    else
    {
        res = core::no_memory;
    }
    
    if (res == core::ok)
    {
        int cc = pthread_mutex_init(&mtx_, NULL);
        if (cc == EOK)
        {
            cc = pthread_cond_init(&cond_, NULL);
            if (cc == EOK)
            {
                wake_up_ = false;
                waiting_on_pipe_ = false;
            }
            else
            {
                (void)pthread_mutex_destroy(&mtx_);
            }
        }
        
        if (cc != EOK)
        {
            ::close(interrupt_pipe_write_);
            ::close(interrupt_pipe_read_);
            
            alloc_->deallocate(polls_);
            
            res = core::io_error;
        }
    }

    return res;
}

void selector::install_io_error_logger(
    core::io_error_function io_error_callback,
    void * io_error_callback_hint)
{
    io_error_callback_ = io_error_callback;
    io_error_callback_hint_ = io_error_callback_hint;
}

void selector::clean()
{
    ::close(interrupt_pipe_write_);
    ::close(interrupt_pipe_read_);

    if (polls_ != NULL)
    {
        alloc_->deallocate(polls_);
        polls_ = NULL;
    }
    
    (void)pthread_cond_destroy(&cond_);
    (void)pthread_mutex_destroy(&mtx_);
}

void selector::reset()
{
    // always add the internal interrupt pipe

    polls_[0].fd = interrupt_pipe_read_;
    polls_[0].events = POLLIN;

    polls_used_ = 1;

#ifdef YAMI4_WITH_OPEN_SSL
    ssl_pending_ = false;
#endif // YAMI4_WITH_OPEN_SSL
}

core::result selector::grow_polls()
{
    core::result res;

    std::size_t new_size = 2 * polls_size_;

    pollfd * new_array = static_cast<pollfd *>(
        alloc_->allocate(sizeof(struct pollfd) * new_size));
    if (new_array != NULL)
    {
        std::memcpy(new_array, polls_, sizeof(struct pollfd) * polls_size_);

        alloc_->deallocate(polls_);

        polls_ = new_array;
        polls_size_ = new_size;

        res = core::ok;
    }
    else
    {
        res = core::no_memory;
    }

    return res;
}

core::result selector::add_channel(channel & ch,
    bool allow_outgoing_traffic, bool allow_incoming_traffic)
{
    core::result res = core::ok;
    
    // QNX channels are not registered in the selector,
    // they are processed directly in a separate step of the worker
    
    protocol proto = ch.get_protocol();
    if (proto != proto_qnx)
    {
        // file-descriptor-based channels are registered
        // in the selector for further polling
        
        if (polls_used_ == polls_size_)
        {
            res = grow_polls();
        }

        if (res == core::ok)
        {
            io_descriptor_type fd = ch.get_io_descriptor();
            io_direction direction = ch.get_io_direction();
            
            pollfd & pf = polls_[polls_used_];
            pf.events = 0;
            pf.revents = 0;

#ifdef YAMI4_WITH_OPEN_SSL
            ch.set_pending_read(false);
#endif // YAMI4_WITH_OPEN_SSL

            if ((direction == input || direction == inout) && allow_incoming_traffic)
            {
                pf.fd = fd;
                pf.events |= POLLIN;

#ifdef YAMI4_WITH_OPEN_SSL
                SSL * ssl = ch.get_ssl();
                if (ssl != NULL)
                {
                    int pending = SSL_pending(ssl);
                    if (pending != 0)
                    {
                        ssl_pending_ = true;

                        ch.set_pending_read(true);
                    }
                }
#endif // YAMI4_WITH_OPEN_SSL
            }

            if ((direction == output || direction == inout) && allow_outgoing_traffic)
            {
                pf.fd = fd;
                pf.events |= POLLOUT;
            }

            ch.set_selector_index(polls_used_);
            
            ++polls_used_;
        }
    }

    return res;
}

core::result selector::add_listener(listener & lst, bool allow_incoming_traffic)
{
    core::result res = core::ok;

    // QNX listeners are not registered in the selector,
    // but instead actively interrupt the selector when
    // they have some new message frame to process
    
    if (allow_incoming_traffic)
    {
        protocol proto = lst.get_protocol();
        if (proto != proto_qnx)
        {
            // file-descriptor-based listeners are registered
            // in the selector for further polling
            
            if (polls_used_ == polls_size_)
            {
                res = grow_polls();
            }

            if (res == core::ok)
            {
                io_descriptor_type fd = lst.get_io_descriptor();

                pollfd & pf = polls_[polls_used_];
                pf.fd = fd;
                pf.events = POLLIN;
                pf.revents = 0;

                lst.set_selector_index(polls_used_);
            
                ++polls_used_;
            }
        }
    }

    return res;
}

void selector::get_channel_usage(std::size_t & max_allowed, std::size_t & used)
{
    max_allowed = 0;
    used = polls_used_;
}

core::result selector::wait(std::size_t timeout)
{
    core::result res = core::ok;

#ifdef YAMI4_WITH_OPEN_SSL
    if (ssl_pending_)
    {
        // if there are any bytes pending in SSL buffers,
        // we do not want to wait on select, but still want
        // to poll the state of all file descriptors

        timeout = 0;
    }
#endif // YAMI4_WITH_OPEN_SSL

    int cc = pthread_mutex_lock(&mtx_);
    if (cc != EOK)
    {
        res = core::io_error;
    }
    
    if (wake_up_ == false)
    {
        if (polls_used_ == 1)
        {
            // the poll array contains only the (initial) entry
            // for the internal pipe, meaning that
            // no file-descriptor-based channels were registered
            // and the only channels (if any) that are in use are QNX channels
            // -> wait on condvar
            
            // Note: this scheme gives preference to QNX channels
            // and it is possible for busy QNX communication
            // to starve file-descriptor-based channels, as they might never
            // get checked for readiness.
            
            timespec tp;
            cc = clock_gettime(CLOCK_REALTIME , &tp);
            if (cc == 0)
            {
                nsec2timespec(&tp,
                    timespec2nsec(&tp) +
                        static_cast<uint64_t>(timeout) * 1000000u
                );
            }
            else
            {
                res = core::io_error;
            }
            
            while (res == core::ok && wake_up_ == false)
            {
                cc = pthread_cond_timedwait(&cond_, &mtx_, &tp);
                if (cc != EOK)
                {
                    if (cc == ETIMEDOUT)
                    {
                        res = core::timed_out;
                    }
                    else
                    {
                        res = core::io_error;
                    }
                }
            }
        }
        else
        {
            // there are some file-descriptor-based channels
            // registered in the poll array
            // -> wait for poll
            
            waiting_on_pipe_ = true;
        }
    }
    
    // prepare for future waits
    wake_up_ = false;
    
    cc = pthread_mutex_unlock(&mtx_);
    if (cc != EOK)
    {
        res = core::io_error;
    }
    
    if (res == core::ok && waiting_on_pipe_)
    {
        int ccp = ::poll(polls_, polls_used_, static_cast<int>(timeout));
        
        // Note:
        // There is a window of opportunity here, between call to ::poll above
        // and clearing the waiting_on_pipe_ flag below,
        // when other threads might try to interrupt the selector
        // by writing to the pipe (because other threads still see
        // waiting_on_pipe_ == true here);
        // this attempt to interrupt the selector
        // will not be noticed by the ::poll function, which has just finished.
        // This hazard is not dangerous, because the selector is awaken anyway
        // (for whatever reason) and further processing will find
        // the new state of the system as expected.
        
        cc = pthread_mutex_lock(&mtx_);
        if (cc != EOK)
        {
            res = core::io_error;
        }
        
        waiting_on_pipe_ = false;
        
        cc = pthread_mutex_unlock(&mtx_);
        if (cc != EOK)
        {
            res = core::io_error;
        }
        
        if (res == core::ok)
        {
            if (ccp > 0)
            {
                // the wait might have been interrupted with the internal pipe
                // if that is the case, consume the dummy info

                // the internal pipe always has the first slot in the polls_ array

                if ((polls_[0].revents & POLLIN) != 0)
                {
                    bool keep_trying = true;
                    while (keep_trying)
                    {
                        char dummy;
                        ssize_t rn =
                            ::read(interrupt_pipe_read_, &dummy, sizeof(dummy));

                        if (rn == -1)
                        {
                            // if the attempt to read is abandoned due to signal,
                            // it should be repeated
                            // otherwise it was an error

                            if (errno != EINTR)
                            {
                                handle_io_error("selector pipe read",
                                    io_error_callback_, io_error_callback_hint_);

                                res = core::io_error;
                                keep_trying = false;
                            }
                        }
                        else
                        {
                            res = core::ok;
                            keep_trying = false;
                        }
                    }
                }
                else
                {
                    // the wait was interrupted because there is
                    // a regular event to process

                    res = core::ok;
                }
            }
            else if (ccp == 0)
            {
#ifdef YAMI4_WITH_OPEN_SSL
                if (ssl_pending_)
                {
                    // the timeout == 0 expired, but since there are some
                    // pending bytes in SSL buffers, we consider it as
                    // a valid outcome of the selection process

                    res = core::ok;
                }
                else
                {
                    res = core::timed_out;
                }
#else // YAMI4_WITH_OPEN_SSL

                res = core::timed_out;

#endif // YAMI4_WITH_OPEN_SSL
            }
            else
            {
                if (ccp == EINTR)
                {
                    // the wait operation will be repeated in the outer loop
                    
                    res = core::ok;
                }
                else
                {
                    handle_io_error("poll wait",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                }
            }
        }
    }

    return res;
}

bool selector::is_channel_ready(
    const channel & ch, io_direction & direction) const
{
    bool ready_for_reading = false;
    bool ready_for_writing = false;

    int index = ch.get_selector_index();

    // Note: QNX channels are not registered in the selector,
    // and retain their default index == -1.
    if (index > 0)
    {
        io_direction dir = ch.get_io_direction();
        
        pollfd & pf = polls_[index];
        
        if (dir == input || dir == inout)
        {
            if ((pf.revents & POLLIN) != 0)
            {
                ready_for_reading = true;
            }
            
#ifdef YAMI4_WITH_OPEN_SSL
            if (ch.get_pending_read())
            {
                ready_for_reading = true;
            }
#endif // YAMI4_WITH_OPEN_SSL
        }
        
        if (dir == output || dir == inout)
        {
            if ((pf.revents & POLLOUT) != 0)
            {
                ready_for_writing = true;
            }
        }
        
        if (ready_for_reading && ready_for_writing)
        {
            direction = inout;
        }
        else if (ready_for_reading)
        {
            direction = input;
        }
        else if (ready_for_writing)
        {
            direction = output;
        }
    }

    return ready_for_reading || ready_for_writing;
}

bool selector::is_listener_ready(const listener & lst) const
{
    bool ready = false;
    
    protocol proto = lst.get_protocol();
    if (proto != proto_qnx)
    {
        int index = lst.get_selector_index();
        
        if (index > 0)
        {
            pollfd & pf = polls_[index];
            
            ready = (pf.revents & POLLIN) != 0;
        }
    }
    else
    {
        ready = lst.is_qnx_receiver_ready();
    }
    
    return ready;
}

core::result selector::wake_up()
{
    // wake the condvar or poll up from its current wait state

    core::result res = core::ok;
    
    bool wake_up_pipe;
    
    int cc = pthread_mutex_lock(&mtx_);
    if (cc != EOK)
    {
        res = core::io_error;
    }
    
    if (waiting_on_pipe_)
    {
        wake_up_pipe = true;
    }
    else
    {
        wake_up_pipe = false;
        
        // wake up the condvar, as only QNX channels are involved
        wake_up_ = true;
        cc = pthread_cond_signal(&cond_);
        if (cc != EOK)
        {
            res = core::io_error;
        }
    }
    
    cc = pthread_mutex_unlock(&mtx_);
    if (cc != EOK)
    {
        res = core::io_error;
    }
    
    if (res == core::ok && wake_up_pipe)
    {
        bool keep_trying = true;
        while (keep_trying)
        {
            char dummy = '\0';
            ssize_t written =
                ::write(interrupt_pipe_write_, &dummy, sizeof(dummy));

            if (written == -1)
            {
                // if the attempt to write is abandoned due to signal,
                // it should be repeated
                // otherwise it was an error

                if (errno != EINTR)
                {
                    handle_io_error("poll pipe write",
                        io_error_callback_, io_error_callback_hint_);

                    res = core::io_error;
                    keep_trying = false;
                }
            }
            else
            {
                res = core::ok;
                keep_trying = false;
            }
        }
    }

    return res;
}
