// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#ifndef YAMICORE_SERIALIZATION_H_INCLUDED
#define YAMICORE_SERIALIZATION_H_INCLUDED

#include <yami4-core/allocator.h>
#include <yami4-core/core.h>
#include <cstddef>

namespace yami
{

namespace details
{

inline std::size_t round_up_4(std::size_t v)
{
    return (v + 3) & ~0x03;
}

core::result put_integer(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    int value);

core::result get_integer(const char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    int & value);

core::result put_long_long(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    long long value);

core::result get_long_long(const char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    long long & value);

core::result put_unsigned_integer(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    unsigned int value);

core::result get_unsigned_integer(const char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    unsigned int & value);

core::result put_unsigned_long_long(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    unsigned long long value);

core::result get_unsigned_long_long(const char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    unsigned long long & value);

core::result put_double_float(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    double value);

core::result get_double_float(const char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    double & value);

core::result put_string(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const char * value, std::size_t length);

// optimization:
// if the length of the retrieved string is less than initial length,
// there is no memory allocated and string is written into value directly
core::result get_string(const char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    char * & value, std::size_t & length,
    core::allocator & alloc);

core::result put_boolean_array(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const bool * values, std::size_t length);

core::result get_boolean_array(const char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    bool * & values, std::size_t & length,
    core::allocator & alloc);

core::result put_integer_array(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const int * values, std::size_t length);

core::result get_integer_array(const char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    int * & values, std::size_t & length,
    core::allocator & alloc);

core::result put_long_long_array(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const long long * values, std::size_t length);

core::result get_long_long_array(const char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    long long * & values, std::size_t & length,
    core::allocator & alloc);

core::result put_unsigned_integer_array(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const unsigned int * values, std::size_t length);

core::result get_unsigned_integer_array(const char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    unsigned int * & values, std::size_t & length,
    core::allocator & alloc);

core::result put_unsigned_long_long_array(char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const unsigned long long * values, std::size_t length);

core::result get_unsigned_long_long_array(const char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    unsigned long long * & values, std::size_t & length,
    core::allocator & alloc);

core::result put_double_float_array(char * * buffers,
    const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, char * & buffer_position,
    const double * values, std::size_t length);

core::result get_double_float_array(
    const char * * buffers, const std::size_t * sizes,
    std::size_t num_of_buffers,
    std::size_t & current_buffer, const char * & buffer_position,
    double * & values, std::size_t & length,
    core::allocator & alloc);

void fill_outgoing_frame_header(
    char * buffer,
    std::size_t message_id, int frame_number,
    std::size_t message_header_size, std::size_t frame_payload_size);

} // namespace details

} // namespace yami

#endif // YAMICORE_SERIALIZATION_H_INCLUDED
