// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-core/allocator.h>
#include <algorithm>
#include <cstdlib>
#include <ctime>
#include <cassert>
#include <random>
#include <vector>

// series of allocs, out of order dealloc
void test1()
{
    const std::size_t private_area_size = 80000;
    char area[private_area_size];

    yami::core::non_locked_allocator a;
    a.set_working_area(area, private_area_size);

    const int num_of_blocks = 1000;
    std::vector<void *> blocks;
    for (int i = 0; i != num_of_blocks; ++i)
    {
        void * p = a.allocate(10);
        assert(p != NULL);
        if (blocks.empty() == false)
        {
            // check that the blocks are ordered in the working area
            assert(p > blocks[blocks.size() - 1]);
        }
        blocks.push_back(p);
    }

    std::random_device rd;
    std::mt19937 g(rd());
    std::shuffle(blocks.begin(), blocks.end(), g);

    for (int i = 0; i != num_of_blocks; ++i)
    {
        a.deallocate(blocks[i]);
    }
}

// mixed and out of order alloc/dealloc
void test2()
{
    char area[100000];

    yami::core::non_locked_allocator a;
    a.set_working_area(area, 100000);

    const int num_of_steps = 1000000;
    std::vector<void *> blocks;

    std::srand(static_cast<unsigned int>(std::time(NULL)));
    for (int i = 0; i != num_of_steps; ++i)
    {
        enum action { alloc, dealloc };
        action to_do;

        std::size_t current_blocks = blocks.size();

        if (current_blocks < 100)
        {
            to_do = alloc;
        }
        else if (current_blocks == 1000)
        {
            to_do = dealloc;
        }
        else
        {
            if ((std::rand() & 1) == 0)
            {
                to_do = alloc;
            }
            else
            {
                to_do = dealloc;
            }
        }

        if (to_do == alloc)
        {
            void * p = a.allocate(10);
            assert(p != NULL);
            blocks.push_back(p);
        }
        else
        {
            std::size_t index = std::rand() % current_blocks;
            a.deallocate(blocks[index]);
            blocks.erase(blocks.begin() + index);
        }
    }

    // deallocate remaining blocks
    for (std::size_t i = 0; i != blocks.size(); ++i)
    {
        a.deallocate(blocks[i]);
    }
}

int main()
{
    test1();
    test2();
}
