// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include "selector.h"
#include <yami4-core/allocator.h>
#include "../../channel.h"
#include "../../io_error_handler.h"
#include "../../listener.h"
#include <cstring>

#include <Winsock2.h>

using namespace yami;
using namespace details;

core::result selector::init(core::allocator & alloc)
{
    io_error_callback_ = NULL;
    
    alloc_ = &alloc;
    polls_ = NULL;

    core::result res = core::io_error;

    WORD versionRequested = MAKEWORD(2, 0);
    WSADATA wsadata;
    (void)WSAStartup(versionRequested, &wsadata);

    polls_size_ = 10; // initial size of poll array
    polls_ = static_cast<WSAPOLLFD *>(
        alloc_->allocate(sizeof(*polls_) * polls_size_));
    if (polls_ != NULL)
    {
        polls_used_ = 0;
        
        // create artificial pipe on a pair of sockets

        io_descriptor_type ld = ::socket(AF_INET, SOCK_STREAM, 0);
        if (ld != INVALID_SOCKET)
        {
            int localhost = htonl(0x7f000001); // 127.0.0.1

            sockaddr_in local;

            std::memset(&local, 0, sizeof(local));

            local.sin_family = AF_INET;
            local.sin_port = 0; // system-assigned port
            local.sin_addr.s_addr = localhost;

            int cc = ::bind(ld,
                reinterpret_cast<sockaddr *>(&local), sizeof(local));
            if (cc != SOCKET_ERROR)
            {
                cc = ::listen(ld, 1);
                if (cc != SOCKET_ERROR)
                {
                    // figure out the system-assigned port
                    int dummy = sizeof(local);
                    cc = ::getsockname(ld,
                        reinterpret_cast<sockaddr*>(&local), &dummy);
                    if (cc != SOCKET_ERROR)
                    {
                        const u_short assigned_listener_port = local.sin_port;

                        io_descriptor_type sock1 =
                            ::socket(AF_INET, SOCK_STREAM, 0);
                        if (sock1 != INVALID_SOCKET)
                        {
                            struct sockaddr_in servaddr;

                            std::memset(&servaddr, 0, sizeof(servaddr));
                            servaddr.sin_family = AF_INET;
                            servaddr.sin_port = assigned_listener_port;
                            memcpy(&(servaddr.sin_addr),
                                &localhost, sizeof(int));

                            cc = ::connect(sock1,
                                (struct sockaddr*)&servaddr, sizeof(servaddr));
                            if (cc != SOCKET_ERROR)
                            {
                                // connection created, accept the other end
                                // and close the temporary listening socket

                                sockaddr_in client_addr;
                                int client_addr_size = sizeof(client_addr);

                                io_descriptor_type sock2 = ::accept(ld,
                                    reinterpret_cast<sockaddr *>(&client_addr),
                                    &client_addr_size);
                                if (sock2 != INVALID_SOCKET)
                                {
                                    ::closesocket(ld);

                                    interrupt_pipe_write_ = sock1;
                                    interrupt_pipe_read_ = sock2;

                                    res = core::ok;
                                }
                            }
                        }
                    }
                }
            }
        }
        
        if (res != core::ok)
        {
            alloc_->deallocate(polls_);
            polls_ = NULL;
        }
    }
    else
    {
        res = core::no_memory;
    }

    return res;
}

void selector::install_io_error_logger(
    core::io_error_function io_error_callback,
    void * io_error_callback_hint)
{
    io_error_callback_ = io_error_callback;
    io_error_callback_hint_ = io_error_callback_hint;
}

void selector::clean()
{
    ::closesocket(interrupt_pipe_write_);
    ::closesocket(interrupt_pipe_read_);

    if (polls_ != NULL)
    {
        alloc_->deallocate(polls_);
        polls_ = NULL;
    }
    
    (void)WSACleanup();
}

void selector::reset()
{
#ifdef YAMI4_WITH_OPEN_SSL
    ssl_pending_ = false;
#endif // YAMI4_WITH_OPEN_SSL

    // always add the internal interrupt pipe

    polls_[0].fd = interrupt_pipe_read_;
    polls_[0].events = POLLRDNORM;

    polls_used_ = 1;
}

core::result selector::grow_polls()
{
    core::result res;

    std::size_t new_size = 2 * polls_size_;

    pollfd * new_array = static_cast<WSAPOLLFD *>(
        alloc_->allocate(sizeof(*polls_) * new_size));
    if (new_array != NULL)
    {
        std::memcpy(new_array, polls_, sizeof(*polls_) * polls_size_);

        alloc_->deallocate(polls_);

        polls_ = new_array;
        polls_size_ = new_size;

        res = core::ok;
    }
    else
    {
        res = core::no_memory;
    }

    return res;
}

core::result selector::add_channel(channel & ch,
    bool allow_outgoing_traffic, bool allow_incoming_traffic)
{
    core::result res = core::ok;

    if (polls_used_ == polls_size_)
    {
        res = grow_polls();
    }

    if (res == core::ok)
    {
        io_descriptor_type fd = ch.get_io_descriptor();
        io_direction direction = ch.get_io_direction();

        WSAPOLLFD & pf = polls_[polls_used_];
        pf.events = 0;
        pf.revents = 0;
        
#ifdef YAMI4_WITH_OPEN_SSL
        ch.set_pending_read(false);
#endif // YAMI4_WITH_OPEN_SSL

        if ((direction == input || direction == inout) && allow_incoming_traffic)
        {
            pf.fd = fd;
            pf.events |= POLLRDNORM;

#ifdef YAMI4_WITH_OPEN_SSL
            SSL * ssl = ch.get_ssl();
            if (ssl != NULL)
            {
                int pending = SSL_pending(ssl);
                if (pending != 0)
                {
                    ch.set_pending_read(true);
                    ssl_pending_ = true;
                }
            }
#endif // YAMI4_WITH_OPEN_SSL
        }

        if ((direction == output || direction == inout) && allow_outgoing_traffic)
        {
            pf.fd = fd;
            pf.events |= POLLWRNORM;
        }
        
        ch.set_selector_index(polls_used_);
        
        ++polls_used_;
    }

    return res;
}

core::result selector::add_listener(listener & lst, bool allow_incoming_traffic)
{
    core::result res = core::ok;
    
    if (allow_incoming_traffic)
    {
        if (polls_used_ == polls_size_)
        {
            res = grow_polls();
        }
	
        if (res == core::ok)
        {
            io_descriptor_type fd = lst.get_io_descriptor();
            
            WSAPOLLFD & pf = polls_[polls_used_];
            pf.fd = fd;
            pf.events = POLLRDNORM;
            pf.revents = 0;
            
            lst.set_selector_index(polls_used_);
            
            ++polls_used_;
        }
    }
    
    return res;
}

void selector::get_channel_usage(std::size_t & max_allowed, std::size_t & used)
{
    max_allowed = 0;
    used = polls_used_;
}

core::result selector::wait(std::size_t timeout)
{
    core::result res;

    struct timeval tv;

#ifdef YAMI4_WITH_OPEN_SSL
    if (ssl_pending_)
    {
        // if there are any bytes pending in SSL buffers,
        // we do not want to wait on select, but still want
        // to poll the state of all file descriptors

        timeout = 0;
    }
#endif // YAMI4_WITH_OPEN_SSL

    tv.tv_sec = static_cast<long>(timeout / 1000);
    tv.tv_usec = 1000 * (timeout % 1000);

    const int cc = ::WSAPoll(polls_,
        static_cast<unsigned long>(polls_used_), static_cast<int>(timeout));
    if (cc > 0)
    {
        // the wait might have been interrupted with the internal pipe
        // if that is the case, consume the dummy info

        // the internal pipe always has the first slot in the polls_ array

        if ((polls_[0].revents & POLLRDNORM) != 0)
        {
            char dummy;
            int rn =
                ::recv(interrupt_pipe_read_, &dummy, sizeof(dummy), 0);
            if (rn == SOCKET_ERROR)
            {
                handle_io_error("selector pipe read",
                    io_error_callback_, io_error_callback_hint_);

                res = core::io_error;
            }
            else
            {
                res = core::ok;
            }
        }
        else
        {
            // the wait was interrupted because there is
            // a regular event to process

            res = core::ok;
        }
    }
    else if (cc == 0)
    {
#ifdef YAMI4_WITH_OPEN_SSL
        if (ssl_pending_)
        {
            // the timeout == 0 expired, but since there are some
            // pending bytes in SSL buffers, we consider it as
            // a valid outcome of the selection process

            res = core::ok;
        }
        else
        {
            res = core::timed_out;
        }
#else // YAMI4_WITH_OPEN_SSL

        res = core::timed_out;

#endif // YAMI4_WITH_OPEN_SSL
    }
    else
    {
        handle_io_error("poll wait",
            io_error_callback_, io_error_callback_hint_);

        res = core::io_error;
    }

    return res;
}

bool selector::is_channel_ready(
    const channel & ch, io_direction & direction) const
{
    io_direction dir = ch.get_io_direction();

    bool ready_for_reading = false;
    bool ready_for_writing = false;

    std::size_t index = ch.get_selector_index();

    if (index > 0)
    {
        WSAPOLLFD & pf = polls_[index];
        
        if (dir == input || dir == inout)
        {
            if (((pf.revents & POLLRDNORM) != 0) ||
                ((pf.revents & POLLHUP) != 0))
            {
                // channel is ready for reading or closing
                ready_for_reading = true;
            }

#ifdef YAMI4_WITH_OPEN_SSL
            if (ch.get_pending_read())
            {
                ready_for_reading = true;
            }
#endif // YAMI4_WITH_OPEN_SSL
        }

        if (dir == output || dir == inout)
        {
            if ((pf.revents & POLLWRNORM) != 0)
            {
                ready_for_writing = true;
            }
        }

        if (ready_for_reading && ready_for_writing)
        {
            direction = inout;
        }
        else if (ready_for_reading)
        {
            direction = input;
        }
        else if (ready_for_writing)
        {
            direction = output;
        }
    }

    return ready_for_reading || ready_for_writing;
}

bool selector::is_listener_ready(const listener & lst) const
{
    bool ready = false;
    
    std::size_t index = lst.get_selector_index();
    
    if (index > 0)
    {
        WSAPOLLFD & pf = polls_[index];
        
        ready = (pf.revents & POLLRDNORM) != 0;
    }
    
    return ready;
}

core::result selector::wake_up()
{
    // wake the poll up from its current wait state

    core::result res = core::ok;

    char dummy = '\0';
    int written =
        ::send(interrupt_pipe_write_, &dummy, sizeof(dummy), 0);
    if (written == SOCKET_ERROR)
    {
        handle_io_error("poll pipe write",
            io_error_callback_, io_error_callback_hint_);

        res = core::io_error;
    }
    else
    {
        res = core::ok;
    }

    return res;
}
