// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#ifndef YAMICPP_AGENT_IMPL_H_INCLUDED
#define YAMICPP_AGENT_IMPL_H_INCLUDED

#include <yami4-cpp/agent_impl_base.h>
#include "incoming_message_info.h"
#include "incoming_message_queue.h"
#include "name_resolver.h"
#include "number_generator.h"
#include "options.h"
#include <yami4-cpp/outgoing_message.h>
#include "outgoing_message_manager.h"
#include "water_flow_manager.h"
#include <yami4-core/agent.h>
#include <map>
#include <string>
#include <vector>

#include <condition_variable>
#include <mutex>

namespace yami
{

class event_callback;
class parameters;
class serializable;

namespace details
{

class incoming_message_dispatcher_base;

class agent_impl : public agent_impl_base
{
public:

    agent_impl(const parameters & options);

    agent_impl(const parameters & options, event_callback & event_listener);

    ~agent_impl();

    std::string add_listener(const std::string & listener);
    void remove_listener(const std::string & listener);

    void open_connection(const std::string & target);
    void open_connection(const std::string & target,
        const parameters & options);

    virtual long long get_next_message_id();

    std::unique_ptr<outgoing_message> send(
        const std::string & target,
        const std::string & object_name,
        const std::string & message_name,
        const serializable & content,
        std::size_t priority,
        bool auto_connect);

    void send(
        outgoing_message & message,
        const std::string & target,
        const std::string & object_name,
        const std::string & message_name,
        const serializable & content,
        std::size_t priority,
        bool auto_connect);

    std::unique_ptr<outgoing_message> send(
        const std::string & target,
        const serializable & header,
        const serializable & content,
        long long message_id,
        std::size_t priority,
        bool auto_connect);

    void send(
        outgoing_message & message,
        const std::string & target,
        const serializable & header,
        const serializable & content,
        long long message_id,
        std::size_t priority,
        bool auto_connect);

    void clean_outgoing_message_callback(long long id);
    
    void send_one_way(const std::string & target,
        const std::string & object_name,
        const std::string & message_name,
        const serializable & content,
        std::size_t priority,
        bool auto_connect);

    void send_one_way(const std::string & target,
        const serializable & header,
        const serializable & content,
        long long message_id,
        std::size_t priority,
        bool auto_connect);

    void send_reply(const std::string & source,
        long long message_id, const serializable & body,
        std::size_t priority);

    void send_rejection(const std::string & source,
        long long message_id, const std::string & reason,
        std::size_t priority);

    void close_connection(const std::string & target, std::size_t priority);

    void hard_close_connection(const std::string & target);

    virtual void register_object(
        const std::string & object_name,
        std::unique_ptr<incoming_message_dispatcher_base> & object);

    void unregister_object(const std::string & object_name);

    virtual long long send(
        std::unique_ptr<outgoing_message_dispatcher_base> &
            outgoing_message_callback,
        const std::string & target,
        const std::string & object_name,
        const std::string & message_name,
        const serializable & content,
        std::size_t priority,
        bool auto_connect);
        
    virtual void send(
        std::unique_ptr<outgoing_message_dispatcher_base> &
            outgoing_message_callback,
        const std::string & target,
        const serializable & header,
        const serializable & content,
        long long message_id,
        std::size_t priority,
        bool auto_connect);

    virtual void register_connection_event_monitor(
        std::unique_ptr<connection_event_dispatcher_base> & monitor);

    virtual void register_io_error_logger(
        std::unique_ptr<io_error_dispatcher_base> & logger);

    void report_connection_event(
        const char * name, connection_event event);

    // executed as a callback when the incoming message arrives
    void queue_incoming(std::unique_ptr<incoming_message_info> & incoming);

    // executed as a callback when the reply arrives
    void report_replied(long long message_id,
        std::unique_ptr<core::parameters> & header, std::unique_ptr<parameters> & body);
    void report_replied(long long message_id,
        std::unique_ptr<core::parameters> & header, std::unique_ptr<std::vector<char> > & raw_buffer);

    // executed as a callback when the rejection/exception arrives
    void report_rejected(long long message_id,
        std::unique_ptr<core::parameters> & header, const std::string & reason);

    // executed from separate thread, dispatches queued incoming messages
    void do_message_dispatching(std::size_t dispatcher_index);

    // executed from dedicated worker thread, runs the I/O event loop
    void do_event_loop();

    void get_outgoing_flow_state(std::size_t & current_level,
        std::size_t & high_water_mark, std::size_t & low_water_mark) const;

    void decrease_outgoing_flow();

    void get_channel_usage(std::size_t & max_allowed, std::size_t & used);
    
    std::size_t get_pending_outgoing_bytes(const std::string & target);
    
    cpp_options options_;

private:
    agent_impl(const agent_impl &);
    void operator=(const agent_impl &);

    void init(const parameters & options);
    void clean();

    enum message_creation_policy
    {
        none,
        create_new,       // new out_msg object is created dynamically

        replace,          // user-provided object is reinitialized in-place

        helper,           // a helper outgoing_message is created to handle
                          // one-way interaction with forced wait

        keep_for_callback // out_msg object is created, stored for callback
                          // and destroyed after the callback is delivered
    };

    std::unique_ptr<outgoing_message> do_send(
        const std::string & target,
        const std::string & object_name,
        const std::string & message_name,
        const core::serializable & content,
        std::size_t priority,
        bool auto_connect,
        bool one_way,
        std::unique_ptr<outgoing_message_dispatcher_base> &
            outgoing_message_callback,
        outgoing_message * message,
        message_creation_policy message_create,
        long long * out_message_id = NULL);

    std::unique_ptr<outgoing_message> do_send(
        const std::string & target,
        const core::serializable & header,
        const core::serializable & content,
        long long message_id,
        std::size_t priority,
        bool auto_connect,
        bool one_way,
        std::unique_ptr<outgoing_message_dispatcher_base> &
            outgoing_message_callback,
        outgoing_message * message,
        message_creation_policy message_create);

    std::unique_ptr<outgoing_message> do_send_to_single_target(
        const std::string & target,
        const core::serializable & header,
        const core::serializable & content,
        long long message_id,
        std::size_t priority,
        bool auto_connect,
        bool one_way, bool wait_for_transmission, bool wait_for_completion,
        std::unique_ptr<outgoing_message_dispatcher_base> &
            outgoing_message_callback,
        outgoing_message * message,
        message_creation_policy message_create);

    core::channel_descriptor make_sure_channel_exists(
        const std::string & target, bool auto_connect,
        const parameters * overriding_options = NULL);

    void increase_outgoing_flow();
    void increase_incoming_flow();
    void decrease_incoming_flow();

    void stop_worker_thread();
    void stop_dispatcher_threads(std::size_t num_of_threads);

    core::agent agent_;

    outgoing_message_manager outgoing_manager_;

    water_flow_manager outgoing_flow_manager_;
    bool allow_outgoing_flow_;
    std::mutex allow_outgoing_flow_mtx_;
    std::condition_variable allow_outgoing_flow_cond_;

    incoming_message_queue incoming_queue_;
    
    std::vector<bool> dispatcher_stopped_flags_;
    std::mutex dispatcher_stopped_flags_mtx_;
    std::condition_variable dispatcher_stopped_flags_cond_;

    water_flow_manager incoming_flow_manager_;
    bool allow_incoming_traffic_;
    std::mutex allow_incoming_traffic_mtx_;

    std::unique_ptr<connection_event_dispatcher_base> connection_event_monitor_;
    std::unique_ptr<io_error_dispatcher_base> io_error_logger_;
    event_callback * event_listener_;

    name_resolver resolver_;

    number_generator num_gen_;

    bool worker_stop_request_;
    bool worker_stopped_;
    std::mutex worker_stop_mtx_;
    std::condition_variable worker_stop_cond_;
};

} // namespace details

} // namespace yami

#endif // YAMICPP_AGENT_IMPL_H_INCLUDED
