// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-cpp/yami.h>

#include <cstdlib>
#include <iostream>
#include <string>

int main(int argc, char * argv[])
{
    if (argc != 4 && argc != 5)
    {
        std::cout << "expecting three or four parameters: "
            "server destination int_1 int_2 [mode]\n";
        return EXIT_FAILURE;
    }

    const std::string server_address = argv[1];

    int a = std::stoi(argv[2]);
    int b = std::stoi(argv[3]);

    std::string mode;
    bool has_mode = false;
    if (argc == 5)
    {
        mode = argv[4];

        has_mode = true;
    }

    try
    {
        yami::agent client_agent;

        // this step is needed on QNX, to ensure
        // that there exists a possibility to receive responses
        //(void)client_agent.add_listener("qnx://*");

        // this client takes responsibility for preparing the entire message header
        yami::parameters header;
        yami::agent::outgoing_message_id message_id = client_agent.get_next_message_id();
        header.set_string("type", "message");
        header.set_string("object_name", "calculator");
        header.set_string("message_name", "calculate");
        header.set_long_long("message_id", message_id);

        // custom header field:
        if (has_mode)
        {
            header.set_string("mode", mode);
        }

        yami::parameters params;
        params.set_integer("a", a);
        params.set_integer("b", b);

        auto om = client_agent.send(server_address, header, params, message_id);

        om->wait_for_completion();
        const yami::message_state state = om->get_state();
        if (state == yami::replied)
        {
            const yami::parameters & reply =
                om->get_reply();

            int sum = reply.get_integer("sum");
            int difference = reply.get_integer("difference");
            int product = reply.get_integer("product");

            int ratio;
            yami::parameter_entry ratio_entry;
            const bool ratio_defined =
                reply.find("ratio", ratio_entry);
            if (ratio_defined)
            {
                ratio = ratio_entry.get_integer();
            }

            std::cout << "sum        = "
                << sum << '\n';
            std::cout << "difference = "
                << difference << '\n';
            std::cout << "product    = "
                << product << '\n';

            std::cout << "ratio      = ";
            if (ratio_defined)
            {
                std::cout << ratio;
            }
            else
            {
                std::cout << "<undefined>";
            }
        }
        else if (state == yami::rejected)
        {
            std::cout << "The message has been rejected: "
                << om->get_exception_msg();
        }
        else
        {
            std::cout << "The message has been abandoned.";
        }

        std::cout << std::endl;

        if (state == yami::replied || state == yami::rejected)
        {
            // check custom header field in reply

            const yami::parameters & reply_header = om->get_header();
            yami::parameter_entry mode_received_entry;
            if (reply_header.find("mode_received", mode_received_entry))
            {
                bool mode_received = mode_received_entry.get_boolean();
                if (mode_received)
                {
                    std::cout << "the server has received mode setting" << std::endl;
                }
            }
        }
    }
    catch (const std::exception & e)
    {
        std::cout << "error: " << e.what() << std::endl;
    }
}
