// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-cpp/yami.h>

#include <cstdlib>
#include <iostream>
#include <string>

yami::agent * agent_object;

void call(yami::incoming_message & im)
{
    // extract the custom header field named "mode", if it exists

    const yami::parameters & header = im.get_header();

    std::string mode;
    yami::parameter_entry mode_entry;
    if (header.find("mode", mode_entry))
    {
        mode = mode_entry.get_string();
    }

    // extract the parameters for calculations

    const yami::parameters & params = im.get_parameters();

    const int a = params.get_integer("a");
    const int b = params.get_integer("b");

    // prepare the answer with results of four calculations

    yami::parameters reply_params;

    reply_params.set_integer("sum", a + b);
    reply_params.set_integer("difference", a - b);
    reply_params.set_integer("product", a * b);

    // if the ratio cannot be computed,
    // it is not included in the response
    // the client will interpret that fact properly
    if (b != 0)
    {
        reply_params.set_integer("ratio", a / b);
    }

    // server can reply in a basic way:
    //
    // im.reply(reply_params);
    //
    // but it can also use customized header fields to create a message
    // that acts as a first-class response:

    long long message_id = im.get_message_id();
    yami::parameters reply_header;
    reply_header.set_string("type", "reply");
    reply_header.set_long_long("message_id", message_id);
    reply_header.set_boolean("mode_received", mode.empty() == false);

    agent_object->send_one_way(im.get_source(),
        reply_header, reply_params, message_id);

    if (mode == "hex")
    {
        std::cout << "custom header set for hex format" << std::endl;
        std::cout << std::hex;

        std::cout << "got message with parameters "
            << "0x" << a << " and 0x" << b;
    }
    else
    {
        std::cout << std::dec;
        std::cout << "got message with parameters "
            << a << " and " << b;
    }

    std::cout << ", response has been sent back" << std::endl;
}

int main(int argc, char * argv[])
{
    if (argc != 2)
    {
        std::cout
            << "expecting one parameter: server destination\n";
        return EXIT_FAILURE;
    }

    const std::string server_address = argv[1];

    try
    {
        yami::agent server_agent;

        agent_object = &server_agent; // so that message callback can use it

        const std::string resolved_address =
            server_agent.add_listener(server_address);

        std::cout << "The server is listening on "
            << resolved_address << std::endl;

        server_agent.register_object("calculator", call);

        // block
        std::string dummy;
        std::cin >> dummy;
    }
    catch (const std::exception & e)
    {
        std::cout << "error: " << e.what() << std::endl;
    }
}
