// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-cpp/incoming_message.h>
#include "agent_impl.h"
#include "details.h"
#include "incoming_message_info.h"
#include <yami4-core/core.h>

using namespace yami;

incoming_message::incoming_message(details::agent_impl & agent,
    details::incoming_message_info & info)
    : agent_(agent)
{
    source_.swap(info.source);
    object_name_.swap(info.object_name);
    message_name_.swap(info.message_name);
    message_id_ = info.message_id;

    core_header_ = info.core_header;
    header_wrapper_ = NULL;

    params_ = info.body;
    info.body = NULL;
    raw_buffer_ = info.raw_buffer;
    info.raw_buffer = NULL;

    already_used_ = false;
}

incoming_message::incoming_message(incoming_message & other)
    : agent_(other.agent_)
{
    source_.swap(other.source_);
    object_name_.swap(other.object_name_);
    message_name_.swap(other.message_name_);
    message_id_ = other.message_id_;

    core_header_ = other.core_header_;
    other.core_header_ = NULL;
    header_wrapper_ = NULL;

    params_ = other.params_;
    other.params_ = NULL;

    raw_buffer_ = other.raw_buffer_;
    other.raw_buffer_ = NULL;

    already_used_ = other.already_used_;
    other.already_used_ = true;
}

incoming_message::~incoming_message()
{
    delete core_header_;
    delete header_wrapper_;
    delete params_;
    delete raw_buffer_;
}

const std::string & incoming_message::get_source() const
{
    return source_;
}

const std::string & incoming_message::get_object_name() const
{
    return object_name_;
}

const std::string & incoming_message::get_message_name() const
{
    return message_name_;
}

const parameters & incoming_message::get_header() const
{
    // header is lazily-initialized,
    // because it is a high-level wrapper (parameters)
    // for the core header object (core::parameters)

    if (header_wrapper_ == NULL)
    {
        header_wrapper_ = new parameters(core_header_);
    }

    return *header_wrapper_;
}

const parameters & incoming_message::get_parameters() const
{
    if (params_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    return *params_;
}

parameters * incoming_message::extract_parameters()
{
    parameters * p = params_;
    params_ = NULL;
    return p;
}

const std::vector<char> & incoming_message::get_raw_content() const
{
    if (raw_buffer_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    return *raw_buffer_;
}

long long incoming_message::get_message_id() const
{
    return message_id_;
}

void incoming_message::reply(const serializable & body, std::size_t priority)
{
    if (already_used_)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    agent_.send_reply(source_, message_id_, body, priority);
    already_used_ = true;
}

void incoming_message::reject(const std::string & reason,
    std::size_t priority)
{
    if (already_used_)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    agent_.send_rejection(source_, message_id_, reason, priority);
    already_used_ = true;
}
