// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include "options.h"
#include <yami4-cpp/option_names.h>
#include <yami4-cpp/parameters.h>

using namespace yami;
using namespace details;

namespace // unnamed
{

void override_if_defined(const parameters & params, const char * name,
    std::size_t & value)
{
    parameter_entry entry;
    const bool found = params.find(name, entry);
    if (found)
    {
        const parameter_type type = entry.type();
        if (type == integer)
        {
            int tmp = entry.get_integer();
            if (tmp >= 0)
            {
                value = static_cast<std::size_t>(tmp);
            }
        }
    }
}

void override_if_defined(const parameters & params, const char * name,
    bool & value)
{
    parameter_entry entry;
    const bool found = params.find(name, entry);
    if (found)
    {
        const parameter_type type = entry.type();
        if (type == boolean)
        {
            value = entry.get_boolean();
        }
    }
}

} // unnamed namespace

void cpp_options::init(const parameters & params)
{
    // set default values

    // by default run the worker and dispatcher thread(s),
    // this option is ignored altogether on RTOS operating systems
    run_internal_threads = true;

    // there is just one dispatcher thread
    dispatcher_threads = 1;

    // try to connect 5 times if not successful
    connection_retries = 5;

    // before retrying the connection wait random time
    // between 0 and this number of milliseconds
    connection_retry_delay_spread = 100;

    // there can be at most 100 messages waiting in the
    // outgoing queues and if this limit is reached, the queue blocks
    // and allows more messages only after the backlog is reduced to 20
    outgoing_high_water_mark = 100;
    outgoing_low_water_mark = 20;

    // there can be at most 100 messages waiting for dispatch
    // in the incoming queue and if this limit is reached, the
    // incoming traffic is suppressed until the queue is reduced to 20
    incoming_high_water_mark = 100;
    incoming_low_water_mark = 20;

    // by default the message content is delivered
    // as already deserialized parameters object
    deliver_as_raw_binary = false;

    // by default there is no timeout for failover components
    // note that this might indefinitely interrupt failover operation
    default_failover_timeout = 0;

    // extract user-provided options and override default settings

    override_if_defined(params,
        option_names::run_internal_threads, run_internal_threads);

    override_if_defined(params,
        option_names::dispatcher_threads, dispatcher_threads);

    override_if_defined(params,
        option_names::connection_retries, connection_retries);

    override_if_defined(params,
        option_names::connection_retry_delay_spread,
        connection_retry_delay_spread);

    override_if_defined(params,
        option_names::outgoing_high_water_mark, outgoing_high_water_mark);

    override_if_defined(params,
        option_names::outgoing_low_water_mark, outgoing_low_water_mark);

    if (outgoing_low_water_mark > outgoing_high_water_mark)
    {
        outgoing_low_water_mark = outgoing_high_water_mark;
    }

    override_if_defined(params,
        option_names::incoming_high_water_mark, incoming_high_water_mark);

    override_if_defined(params,
        option_names::incoming_low_water_mark, incoming_low_water_mark);

    if (incoming_low_water_mark > incoming_high_water_mark)
    {
        incoming_low_water_mark = incoming_high_water_mark;
    }

    override_if_defined(params,
        option_names::deliver_as_raw_binary, deliver_as_raw_binary);

    override_if_defined(params,
        option_names::default_failover_timeout, default_failover_timeout);
}
