// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include <yami4-cpp/outgoing_message.h>
#include "details.h"
#include "outgoing_message_info.h"
#include "outgoing_message_manager.h"

#include <chrono>

using namespace yami;

outgoing_message::outgoing_message()
    : manager_(NULL), info_(NULL), owner_of_info_(false)
{
}

outgoing_message::~outgoing_message()
{
    clean();
}

void outgoing_message::clean()
{
    if (owner_of_info_ && manager_ != NULL && info_ != NULL)
    {
        manager_->remove(info_->message_id);
        info_->dec_ref_count();
    }
}

void outgoing_message::reset(
    details::outgoing_message_manager & manager,
    details::outgoing_message_info & info)
{
    clean();

    manager_ = &manager;
    info_ = &info;
    owner_of_info_ = true;

    info_->bump_ref_count();
}

void outgoing_message::disown_info_object()
{
    info_->dec_ref_count();
    owner_of_info_ = false;
}

message_state outgoing_message::get_state() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    return info_->state;
}

message_state outgoing_message::get_state(
    std::size_t & sent_bytes,
    std::size_t & total_byte_count) const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    sent_bytes = info_->sent_bytes;
    total_byte_count = info_->total_byte_count;

    return info_->state;
}

void outgoing_message::wait_for_transmission() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::unique_lock<std::mutex> lock(info_->mtx);
    while (info_->state == posted)
    {
        info_->cond.wait(lock);
    }
}

bool outgoing_message::wait_for_transmission(
    std::size_t relative_timeout) const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::unique_lock<std::mutex> lock(info_->mtx);
    while (info_->state == posted)
    {
        std::cv_status status =
            info_->cond.wait_for(lock, std::chrono::milliseconds(relative_timeout));
        
        if (status == std::cv_status::timeout)
        {
            break;
        }
    }
    
    return info_->state != posted;
}

void outgoing_message::wait_for_completion() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::unique_lock<std::mutex> lock(info_->mtx);
    while (info_->state == posted || info_->state == transmitted)
    {
        info_->cond.wait(lock);
    }
}

bool outgoing_message::wait_for_completion(
    std::size_t relative_timeout) const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::unique_lock<std::mutex> lock(info_->mtx);
    while (info_->state == posted || info_->state == transmitted)
    {
        std::cv_status status =
            info_->cond.wait_for(lock, std::chrono::milliseconds(relative_timeout));
        
        if (status == std::cv_status::timeout)
        {
            break;
        }
    }
    
    return info_->state == abandoned ||
        info_->state == replied ||
        info_->state == rejected;
}

const parameters & outgoing_message::get_header() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    if (info_->core_reply_header == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    // header wrapper is lazily initialized
    if (info_->header_wrapper == NULL)
    {
        info_->header_wrapper = new parameters(info_->core_reply_header);
    }

    return *(info_->header_wrapper);
}

const parameters & outgoing_message::get_reply() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    if (info_->reply_body == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    return *(info_->reply_body);
}

const std::vector<char> & outgoing_message::get_raw_reply() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    if (info_->reply_raw_buffer == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    return *(info_->reply_raw_buffer);
}

parameters * outgoing_message::extract_reply()
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    parameters * result = info_->reply_body;
    info_->reply_body = NULL;

    return result;
}

const std::string & outgoing_message::get_exception_msg() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    return info_->exception_msg;
}

outgoing_message::outgoing_message_id outgoing_message::get_message_id() const
{
    if (info_ == NULL)
    {
        details::translate_result_to_exception(core::bad_state);
    }

    std::lock_guard<std::mutex> lock(info_->mtx);

    return info_->message_id;
}

