// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include "outgoing_message_manager.h"
#include "outgoing_message_info.h"

using namespace yami;
using namespace details;

outgoing_message_manager::~outgoing_message_manager()
{
    for (auto & e : map_)
    {
        outgoing_message_info * outgoing = e.second;
        e.second = NULL;

        outgoing->dec_ref_count();
    }
}

void outgoing_message_manager::put(long long message_id,
    outgoing_message_info * outgoing)
{
    std::lock_guard<std::mutex> lock(mtx_);

    map_[message_id] = outgoing;
    
    outgoing->bump_ref_count();
}

void outgoing_message_manager::remove(long long message_id)
{
    std::lock_guard<std::mutex> lock(mtx_);

    map_type::iterator it = map_.find(message_id);

    // it is acceptable to have a spurious remove - this can be a result
    // of combination of exception handling and regular cleanup
    if (it != map_.end())
    {
        do_remove(it);
    }
}

void outgoing_message_manager::do_remove(const map_type::iterator & it)
{
    outgoing_message_info * outgoing = it->second;
    
    it->second = NULL;

    outgoing->dec_ref_count(/* stop_further_callbacks = */ true);

    map_.erase(it);
}

void outgoing_message_manager::report(long long message_id, std::unique_ptr<core::parameters> & header,
    std::unique_ptr<parameters> * body,                // only one
    std::unique_ptr<std::vector<char> > * raw_buffer,  // of these three
    const std::string * reason)                        // is not-NULL
{
    map_type::iterator it;
    outgoing_message_info * outgoing = NULL;
    bool state_changed = false;
    bool should_remove = false;
    
    {
        std::lock_guard<std::mutex> lock(mtx_);
        
        it = map_.find(message_id);
        
        // it is not a bug if the user tries to operate on non-existing message
        // (it might be a network junk that should be ignored)
        
        if (it != map_.end())
        {
            outgoing = it->second;
            
            {
                std::lock_guard<std::mutex> lock_outgoing(outgoing->mtx);

                // it is possible to get several replies to the same message
                // (this might be a junk network content),
                // but only the first reply is taken into account

                if (outgoing->state == posted || outgoing->state == transmitted)
                {
                    outgoing->core_reply_header = header.release();

                    if (body != NULL)
                    {
                        // reply with parameters body
                        
                        outgoing->state = replied;

                        outgoing->reply_body = body->release();
                    }
                    else if (raw_buffer != NULL)
                    {
                        // reply with raw buffer
                        
                        outgoing->state = replied;

                        outgoing->reply_raw_buffer = raw_buffer->release();
                    }
                    else
                    {
                        // rejection
                        
                        outgoing->state = rejected;

                        outgoing->exception_msg = *reason;
                    }
                    
                    // required callback
                    state_changed = true;
                    
                    // there will be no more interaction with this message
                    should_remove = true;
                }
            }
        }
    }
    
    if (state_changed)
    {
        outgoing->process_callback();
    }
    
    if (should_remove)
    {
        // there might be user threads
        // waiting for transmitted or completed state,
        // which was just set above,
        // notify just in case - there will be no more opportunity to do so
        outgoing->cond.notify_all();

        std::lock_guard<std::mutex> lock(mtx_);

        // find again, the entry might have disappeared in the meantime
        // due to destruction of the outgoing_message object
        it = map_.find(message_id);
        if (it != map_.end())
        {
            do_remove(it);
        }
    }
}

void outgoing_message_manager::report_replied(long long message_id,
    std::unique_ptr<core::parameters> & header, std::unique_ptr<parameters> & body)
{
    report(message_id, header, &body, NULL, NULL);
}

void outgoing_message_manager::report_replied(long long message_id,
    std::unique_ptr<core::parameters> & header, std::unique_ptr<std::vector<char> > & raw_buffer)
{
    report(message_id, header, NULL, &raw_buffer, NULL);
}

void outgoing_message_manager::report_rejected(long long message_id,
    std::unique_ptr<core::parameters> & header, const std::string & reason)
{
    report(message_id, header, NULL, NULL, &reason);
}
