// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#include "broker.h"
#include "broker-configuration.h"
#include "broker-messaging.h"
#include "broker-routing.h"
#include "../common/log.h"
#include "../common/utils.h"

#include <cstdio>

#include <chrono>
#include <thread>

int main(int argc, char * argv[])
{
    try
    {
        bool success = broker::init_config(argc, argv);

        if (success)
        {
            logger::set_output_file(
                broker::log_file_name(),
                broker::log_file_max_size(),
                broker::log_file_line_sync());
            
            for (logger::module m = logger::main; m != logger::last_module_;
                 m = static_cast<logger::module>(m + 1))
            {
                logger::enable(m, broker::log_enabled(m));
            }

            logger::put(logger::main, "initialized configuration settings");

            broker::init_routing(
                broker::max_subscriptions(),
                broker::max_client_queue(),
                broker::subscription_overflow_policy());

            // install all forward channels

            int forward_index = 1;
            while (true)
            {
                const std::string & target = broker::forward_target(forward_index);
                const std::string & filter = broker::forward_filter(forward_index);

                if (target.empty())
                {
                    break;
                }

                broker::set_forwarding(filter, target);

                ++forward_index;
            }
            
            // take into account run-time options as defined in configuration
            yami::core::parameters options;
            broker::agent_options(options);
            
            broker::init_messaging(options,
                broker::listener(),
                broker::announce_file(),
                broker::subscription_overflow_policy());

            std::this_thread::sleep_for(std::chrono::milliseconds(broker::warmup_time()));

            broker::allow_incoming();
        }
        else
        {
            logger::put(logger::main, "configuration not initialized properly");
        }

        while (true)
        {
            std::this_thread::sleep_for(std::chrono::seconds(10)); // arbitary delay
        }
    }
    catch (const std::exception & e)
    {
        logger::put(logger::main, "unknown exception in main task:");
        logger::put(logger::main, e.what());
    }
}
